#!/usr/bin/perl -w

# Flowcode Custom_Code.c auto-completion-script that makes life easier with custom components
#	written by axon
#
# history:	1.0		09/02/2008	initial release 
#			1.0.1	10/02/2008	fixed a few regex that were a bit too agressive, like neglecting MIP setting when not being <num>=<num>
#			1.1b	22/02/2008	fixed braces style at function start, now it SHOULD work with every style. I do not guarantee it...
#								fixed braces style at the end of a function, now almost everything possible
#			1.2		23/02/2008	major release, because now every style of bad/good coding is supported
#								now backup is saved as Custom_Code.c.<weekday>_<month>_<day>_<hour>-<min>-<sec>_<year> instead of UNIX time
#
# what it does:
# - never deletes your original file, makes a backup every time you call the script
# - filling out MacroNames /-Return / Parameters /-ParamTypes
# - setting the Macro Start/End in each function accordingly
# - working both in Windows and Unix environment (built on a Mac ;-) ) maybe one day we'll see a Linux/Unix/Mac release of FC
# - the MacroIsPrivate settings aren't changed; maybe deleted or added with a preset of 0 for new functions
#
# what it does NOT:
# - correct any code related bugs, use Flowcode debug logging or other compilers
#
# what you need to do:
# - make functions look like this one (not mandatory, but I encourage you to do so):
#
#	void Test(short sBeam, char cFoo){
#		some code;
#		more code;
#	}
#	
# - use only following types in function declarations you want to implement as a "Custom Component":
#	'char' 'char*' 'short' and 'void'
#
# - download ActivePerl from activestate.com for Windows and install it
# - place this script in the same directory as the Custom_Code.c, usually sth like C:\Programs\MatrixMultimedia\Flowcode\Components\
# - run it via command line i.e. "perl set_code.pl" and hit enter. done!
#
#
#	todo (many things ;-) but may take a while):
#	make the backup/copy/move ONLY if "work is done well", and I have no idea how to implement this right now
#	make my code look a bit nicer (will I change that ever???)
#
# BUT BE AWARE THAT THIS SCRIPT IS GIVEN AS IS,
# SO DO NOT MAKE ME LIABLE FOR ANY DISASTEROUS THINGS THAT MIGHT HAPPEN WITH IT.
# YOU HAVE BEEN WARNED!!!
#
# anyway, I hope this makes life easier for you, as it does really for me now :-) and keeps the focus on microC-programming.
#
#

use strict;
use warnings;

# for now it is not an option to give in command line parameters so set_code.pl needs to be in Flowcode/Components/ directory
open(C_C, "<Custom_Code.c") or die "Cannot open file: $!";
my @fct_declaration;
my @types = qw/void char short/;
push @types, quotemeta("char*");
# save the whole "old" Custom_Code.c in an array line for line for later editing
my @wholeFile;

# define pointers to different locations in Custom_Code.c
my $MacroInsertStart;
my $MacroInsertEnd = 10000;
my $GetDefinesEnd;
# the overall function declaration and variable stuff is stored in @ALL
my @ALL;
# array that stores the old MacroIsPrivate section
my @MacroIsPrivateOld;
# because we don't like to scan selfmade functions we have to fill GetDefines in ourself
push @ALL, ["void", "GetDefines", "void GetDefines(){"];

# file/linepointer
my $codeline = 0;

LINE: while(<C_C>){
	
	# save the whole file in an array
	push @wholeFile, $_;
	# save lines which are of interest in a variable
	$MacroInsertStart = $codeline if /\[MacroNames\]/;
	push @MacroIsPrivateOld, "\n$_" if/\[MacroIsPrivate\]/;
	# save old MacroIsPrivate settings
	if(/^\s*\d\s*=\s*\d/ && @MacroIsPrivateOld && $codeline < $MacroInsertEnd){
		# eliminate whitespaces
		s/\t| //g;
		push @MacroIsPrivateOld, $_;
	}
	$MacroInsertEnd = $codeline-6 if /\/\*InitialisationCode_Start\*\//;
	$GetDefinesEnd = $codeline if /\/\*Macro_GetDefines_End\*\//;
	
	# increment file/linepointer
	$codeline++;
	# start looking for other functions as soon, as GetDefines ends
	next unless $GetDefinesEnd;
	foreach my $type (@types){
		# search for the pattern of a function declaration
		if (/($type)\s+(\w+)\s*?\((.*)\)/i){
			# save the backreferences of the regexp if existent
			my $fctFull = "$&";
			my $fctType = $type;
			my $fctName = $2;
			my $fctVars = $3;
			# now get rid of commas, tabs and repeated whitespaces
			$fctVars =~ s/,//g;
			$fctVars =~ s/ +/ /g;
			$fctVars =~ s/\t+//g;
			my @fctVarsArray = split (/ /,$fctVars);
			# debug printing
			#print $fctType," ",$fctName," ",$fctVars,"\n"; 
			# save an anonymous reference of the function declaration inside @ALL, reachable via @ALL[0][0] = fct type, ...
			push @ALL, [$fctType, $fctName, $fctFull,(scalar @fctVarsArray)/2,@fctVarsArray,"\n"];
			# goto lext line of Custom_code.c
			next LINE;
		}
	}
}

# close original Custom_Code.c filehandle
close C_C;

# building arrays that are going to replace the Macro definitions inside Custom_Code.c
my @MacroNamesInsert;
my $count = scalar @ALL;
push @MacroNamesInsert, "[MacroNames]\nCount=$count\n"; 
my @MacroReturnsInsert;
push @MacroReturnsInsert, "\n\n[MacroReturns]\n";

my @MacroParamInsert;
my $MacroParamSubString;
my $MacroParamTypesSubString;

for(my $i = 0;$i < scalar @ALL; $i++){
	my $number = $i + 1;
	push @MacroNamesInsert, "$number=$ALL[$i][1]\n";
	push @MacroReturnsInsert, "$number=$ALL[$i][0]\n";
	# substrings for parameters and types
	$MacroParamSubString = "\n\n[MacroParameters_$ALL[$i][1]]\n";
	$MacroParamTypesSubString = "\n[MacroParamTypes_$ALL[$i][1]]\n"; 
	if(!$ALL[$i][3]){
		$MacroParamSubString .= "Count=0\n";
	}
	else{
		$MacroParamSubString .= "Count=$ALL[$i][3]\n";
		for(my $j = 1; $j <= $ALL[$i][3];$j++){
			$MacroParamSubString .= "$j=$ALL[$i][3+2*$j]\n";
			$MacroParamTypesSubString .= "$j=$ALL[$i][2+2*$j]\n";
		} 	
	}
	push @MacroParamInsert, ($MacroParamSubString, $MacroParamTypesSubString);
}

# bad way of finding out if curly braces are somewhere to find. flag is set, if not behind function declaration
my $foundStuff = 0;
my $CurlyBraces = 0;

# go through whole file
for(my $i = $GetDefinesEnd; $i < scalar @wholeFile;$i++){
	# step through every function's full declaration that is saved in @ALL and compare to active line
	for(my $j=0;$j < scalar @ALL;$j++){
		my $fct = quotemeta($ALL[$j][2]);
		if($wholeFile[$i] =~ /$fct/){

			############
			# compatibility reasons made me code this part: the different braces style of functions ;-)
			# go through the lines until finding one with "{", maybe some code behind it, and set the flag foundStuff
			while(!($wholeFile[$i] =~ /\s*\{.*/)){
				$i++;
				$foundStuff = 1;
			}
			if($wholeFile[$i] =~ /\)\s*?\{(.{2,})/){
				my $cutOut = $1;
				#print "$fct\{\n";
				splice(@wholeFile,$i,1,"$ALL[$j][2]\{\n");
				splice(@wholeFile,$i+1,0,"$cutOut\n");
			}
			
			# now compare again and if there is code stuff behind "{" then shift it a line downwards, makes code look nicer. reset flag
			$wholeFile[$i] =~ /\s*\{(.*)/;
			if($foundStuff && $1 ne ""){
				my $cutOut = $1;
				splice(@wholeFile,$i,1,"{\n");
				splice(@wholeFile,$i+1,0,"$cutOut\n");	
				$foundStuff = 0;
			}
			############
			

			
			# then test if next line already has a Macro_Start and replace it or write a new one
			if($wholeFile[$i+1] =~ /\/\*Macro_/){
				splice(@wholeFile,$i+1,1,"/*Macro_$ALL[$j][1]_Start*/\n");	
			}
			else{
				splice(@wholeFile,$i+1,0,"/*Macro_$ALL[$j][1]_Start*/\n");	
			}
			# now step through the lines until CurlyBraces is zero
			for(my $k = $i;$k < scalar @wholeFile;$k++){
				$CurlyBraces++ if $wholeFile[$k] =~ /\{/g;
				$CurlyBraces-- if $wholeFile[$k] =~ /\}/g;
				next unless $CurlyBraces == 0;
				
				# then, as seen before, look if the "}" has some code before it				
				$wholeFile[$k] =~ /(.*)\}\s*/;
				# if yes, then save the code and set the ending brace alone in one line after the last codeline
				if($1 ne ""){
					my $cutOut = $1;
					splice(@wholeFile,$k,1,"}\n");
					splice(@wholeFile,$k,0,"$cutOut\n");
					$k++;
				}	
								
				if($wholeFile[$k-1] =~ /\/\*Macro_/){
					splice(@wholeFile,$k-1,1,"/*Macro_$ALL[$j][1]_End*/\n");	
				}
				else{
					splice(@wholeFile,$k,0,"/*Macro_$ALL[$j][1]_End*/\n");	
				}						

				# now its easier to go through the rest by setting the wholeFile pointer accordingly and step out
				$i = $k;
				last;
			last;	
			}
		}
		else{
			next;
		}
	}
}
# then lets get the old MacroIsPrivate things working
if($#MacroIsPrivateOld < ($#ALL+1) ){
	for(my $j = $#MacroIsPrivateOld+1; $j <= $#ALL+1; $j++){
		push @MacroIsPrivateOld, "$j=0\n";
	}	
}
if($#MacroIsPrivateOld > ($#ALL+1) ){
	for(my $k = $#MacroIsPrivateOld; $k > $#ALL+1; $k--){
		pop @MacroIsPrivateOld;
		
	}	
}
#print @MacroIsPrivateOld;
#first get out the old Macro definitions
splice(@wholeFile,$MacroInsertStart,$MacroInsertEnd - $MacroInsertStart);
# and then insert the new ones
splice (@wholeFile,$MacroInsertStart,0,(@MacroNamesInsert, @MacroReturnsInsert, @MacroIsPrivateOld, @MacroParamInsert));
#print "@wholeFile";

# changed location of backup if function(s) get(s) caught in endless loops for whatever reason
# localtime in a scalar context returns formatted time
my $now = localtime();
$now =~ s/:/-/g;
$now =~ s/\s/_/g;
# make a backup by moving the old file to one with a time stamp
rename "Custom_Code.c", "Custom_Code.c.$now";

open C_C_NEW, ">Custom_Code.c" or die "Cannot write file";
print C_C_NEW @wholeFile;
close C_C_NEW;
print "Job done! Backup is in Custom_Code.c.$now.\n";